/*********************************************************************
*
*                  PRINT.TTP                  27.12.87
*                  =========
*
*					letzte nderung:	 19.5.89
*
*  Druckprogramm fr den Ausdruck von TOS- Dateien auf
*  Epson- kompatiblen Druckern
*
*  Syntax: print [-p] [-s] [-i] datei1...    - Druckt Datei(en)
*		 -tn							- Tab- Gre
*          -s                                - schmaler Zeilenabstand
*          -p                                - Form Feed nach Ausdruck
*          -i                                - Init vor jeder Datei
*		 -c							- compressed
*		 -d							- double width
*
*********************************************************************/

#include <portab.h>
#include <tos.h>
#include <string.h>


#define ESC    '\33'
#define FF     '\14'
#define CTRL_C '\3'
#define toupper(c)  (c & '\137')

int	tabsize		= 8;
int  form_feed 	= FALSE;
int  small_spc 	= FALSE;
int  to_init   	= FALSE;
int	compressed 	= FALSE;
int  double_width 	= FALSE;


int	drucke(const char *dateiname);
void screen(const char *string);
void print(const char *string);
void toterm(void);



void main(argc,argv)
int argc;
char *argv[];
{
	register int i;
	register char *schalter;


     /* Prfen, ob ohne Parameter gestartet */
	if   (argc < 2)
		{
		message:
		screen("\r\nUsage: PRINT [-psicd] datei1 datei2 ...\r\n\r\n");
		screen("          -tn   Tabgre\r\n"
			  "          -s    Zeilenabstand 1/8\"\r\n"
			  "          -p    Seitenvorschub nach jeder Datei\r\n"
			  "          -i    Init vor jeder Datei\r\n"
			  "          -c    Schmale Zeichen\r\n"
			  "          -d    Breite Zeichen\r\n");
		Pterm(1);
		}

	/* Alle Switches auswerten */
	for  (i = 1; (i < argc) && (argv[i][0] == '-'); i++)
		for	(schalter = argv[i] + 1; *schalter != EOS; schalter++)
			{
			switch(toupper(*schalter))
				{
				case 'T' : {
						 char *endp;

						 tabsize = (int) strtoul(schalter+1, &endp, 10);
						 if	(endp == schalter+1 || tabsize == 0)
						 	{
						 	screen("PRINT: Formatfehler bei -Tn in: ");
						 	goto err;
						 	}
						 schalter = endp-1;
						 }
						 break;
				case 'P' : form_feed    = TRUE;
				           break;
				case 'I' : to_init      = TRUE;
				           break;
				case 'S' : small_spc    = TRUE;
				           break;
				case 'C' : compressed   = TRUE;
				           break;
				case 'D' : double_width = TRUE;
				           break;
				default  : screen("PRINT: Unbekannter Schalter in: ");
						 err:
						 screen(argv[i]);
						 screen("\r\n");
						 Pterm(2);
				}
			}

	/* Dateien drucken */

	if	(i >= argc)				/* keine Dateien */
		goto message;
	for  (; i < argc; i++) {
		if   (to_init)
			{
			toterm();
			print("@");
			}
		if   (small_spc)
			{
			toterm();
			Cprnout(ESC);
			toterm();
			Cprnout('0');
			}
		if   (compressed)
			{
			toterm();
			Cprnout('\x0f');
			}
		if   (double_width)
			{
			toterm();
			Cprnout(ESC);
			toterm();
			Cprnout('W');
			toterm();
			Cprnout('\x01');
			}
		/* Tabgre einstellen */
		toterm();
		Cprnout(ESC);
		toterm();
		Cprnout('e');
		toterm();
		Cprnout(0);
		toterm();
		Cprnout(tabsize);
		if   (drucke(argv[i]))
			if   (form_feed) {
				toterm();
				Cprnout(FF);
				}
		}

	Pterm0();
}


/*********************************************************************
*
*  Druckt die Datei mit Namen <dateiname>
*  Rckgabe FALSE, wenn Fehler
*
*********************************************************************/

int drucke(const char *dateiname)
{
	         long retcode;
	         int  handle;
	         char puffer[2];
	register char *string;


	if   (0L > (retcode = Fopen(dateiname,O_RDONLY))) {
		screen("PRINT: ");
		screen(dateiname);
		screen(" nicht gefunden\r\n");
		return(FALSE);
		}
	handle = (int) retcode;

	screen("Drucke ");
	screen(dateiname);
	screen("\r\n");

     while(1L == Fread(handle,1L,puffer)) {
          puffer[1] = EOS;
          string = puffer;
          switch(puffer[0]) {
               case '#':
               case '$':
               case '@':
               case '[':
               case '\\':
               case ']':
               case '^':
               case '\'':
               case '{':
               case '|':
               case '}':
               case '~': print("R");
                         Cprnout('\0');      break;
               case '': string = "C\b,";    break;
               case '': string = "R\2}";   break;
               case '': string = "R\1{";   break;
               case '': puffer[0] = 'a';    goto circ;
               case '': puffer[0] = 'e';    goto circ;
               case '': puffer[0] = 'i';    goto circ;
               case '': puffer[0] = 'o';    goto circ;
               case '': puffer[0] = 'u';
                         circ:
                         Cprnout(puffer[0]);
                         print("\bR");
                         Cprnout('\0');
                         puffer[0] = '^';    break;
               case '': string = "R\2{";   break;
               case '': string = "R\1@";   break;
               case '': string = "R\4}";   break;
               case '': string = "R\1\\";  break;
               case '': string = "e\bR\1~";break;
               case '': string = "R\1}";   break;
               case '': string = "i\bR\1~";break;
               case '': string = "R\6~";   break;
               case '': string = "R\2[";   break;
               case '': string = "R\4]";   break;
               case '': string = "R\9@";   break;
               case '': string = "R\4{";   break;
               case '': string = "R\4[";   break;
               case '': string = "R\2|";   break;
               case '': string = "R\6|";   break;
               case '': string = "R\1|";   break;
               case '':
               case '': string = "y\bR\1~";break;
               case '': string = "R\2\\";  break;
               case '': string = "R\2]";   break;
               case '': print("c\bR");
                         Cprnout('\0');
                         string[0] = '|';    break;
               case '': string = "R\3#";   break;
               case '': string = "R\8\\";  break;
               case '': string = "R\2~";   break;
               case '': string[0] = 'f';    break;
               case '': string = "a\b'";    break;
               case '': string = "i\b'";    break;
               case '': string = "o\b'";    break;
               case '': string = "u\b'";    break;
               case '': string = "R\7|";   break;
               case '': string = "R\7\\";  break;
               case '': string = "a\b_";    break;
               case '': string = "o\b_";    break;
               case '': string = "R\7]";   break;
               case '': string = "R\7[";   break;
               case '': Cprnout('a');       goto tilde;
               case '': Cprnout('o');
                         tilde:
                         print("\bR");
                         Cprnout('\0');
                         string[0] = '~';    break;
               case '': string = "R\4\\";  break;
               case '': string = "R\4|";   break;
               case '':
               case '': string[0] = 'A';    break;
               case '': string[0] = 'O';    break;
               case '': string = "R\1~";   break;
               case '': string[0] = '\'';   break;
               case '': string = "\0\0\24\147\24\0\0\0\0";
                         goto graph;
               case '': string = "\0\x18\x38\x6f\xc1\xc1\x6f\x38\x18";
                         goto graph;
               case '': string = "\0\x18\x1c\xf6\x83\x83\xf6\x1c\x18";
                         goto graph;
               case '': string = "\0\x3c\x24\x24\xe7\xc3\x66\x3c\x18";
                         goto graph;
               case '': string = "\0\x18\x3c\x66\xc3\xe7\x24\x24\x3c";
                         goto graph;
               case '': string = "\140\220\220\224\224\377\4\0\0";
                         goto graph;
               case '': string = "\176\303\275\245\245\245\303\176\0";
                         goto graph;
               case '': string = "\176\303\275\251\255\265\303\176\0";
                         goto graph;
               case '': string = "\200\360\200\360\200\100\200\360\0";
					graph:
					{
					register int i;


					/* ungerade Spalten drucken */
					/* ------------------------ */
					print("y\376");
					for (i = 0; i < 9; i++)
					     Cprnout((i & 1) ? string[i] : '\0');
					Cprnout('\376');

					/* ein Schritt zurck */
					/* ------------------ */
					Cprnout('\b');

					/* gerade Spalten drucken */
					/* ---------------------- */
					print("y\376");
					for (i = 0; i < 9; i++)
					     Cprnout((i & 1) ? '\0' : string[i]);
					Cprnout('\376');
					}
					string[0] = EOS;    break;
               case '': string[0] = 'Y';    break;
               case '': string = "R\2@";   break;
               case '': print("R");
                         Cprnout('\0');
                         puffer[0] = '^';    break;
               case '': string = "R\2~";   break;
               case '': string = "=\b_";    break;
               case '': string = "+\b_";    break;
               case '': string = ">\b_";    break;
               case '': string = "<\b_";    break;
               case '': string = ":\b-";    break;
               case '': string = "R\1[";   break;

               default:  if   ((string[0] < 0) || (string[0] == '\177'))
                              string[0] = ' ';
               }
          toterm();
          print(string);

          } /* END WHILE */

     Fclose(handle);
     return(TRUE);
}



/*********************************************************************
*
*  Schreibt <string> auf den Bildschirm
*
*********************************************************************/

void screen(const char *string)
{
     Fwrite(-1,(long) strlen(string), string);
}


/*********************************************************************
*
*  Schreibt <string> auf den Drucker
*
*********************************************************************/

void print(const char *string)
{
     while(*string)
          Cprnout(*string++);
}


/*********************************************************************
*
*  Liest Tastaturpuffer und fragt, ob Programm abgebrochen werden soll
*
*********************************************************************/

void toterm(void)
{
     char c;

     do   {
          if   (Bconstat(2)) { /* Wenn Taste gedrckt */
               while(Bconstat(2))
                    if   (CTRL_C == (0xff & Bconin(2)))
                         Pterm0();
               screen("Abbruch (j/n) ? ");
               Bconout(5, c = Bconin(2));
               if   (toupper(c) == 'J')
                    Pterm0();
               screen("\r\n");
               }
          }
     while(!Cprnos());            /* solange Drucker nicht verfgbar */
}
